// IOCntrl.cpp: implementation of the CIOCntrl class.
//
//

#include <math.h>
#include <float.h>
#include <afxmt.h>
#include "stdafx.h"
#include "winpsk.h"
#include "stdafx.h"
#include "Errorcodes.h"
#include "perform.h"
#include "WinPSKDoc.h"
#include "PskDetTbl.h"
#include "IOCntrl.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


#define BUF_SIZE 4096	//size of data chunks to process at a time from
						// the soundcard


// Non Class routine for launching the worker thread

static UINT ThreadLauncher( CIOCntrl* pIO)
{
	pIO->ProcessLoop();
	return 4;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CIOCntrl::CIOCntrl()
{
	m_fProcThreadQuit = FALSE;

	m_pIOBuf = NULL;			//memory pointers
	m_pDispBuf = NULL;
	m_pCpxBuf = NULL;

	m_ErrorCode = 0;

	m_pDec2InPtr = NULL;		//Decimate by 2 FIR pointers
	m_pDec2Que = NULL;

	m_pSoundDev = NULL;			//object pointers
	m_pPSKMod = NULL;
	m_pPskDet = NULL;
	m_pDoc = NULL;
	m_pProcThread = NULL;
	m_hComPort = INVALID_HANDLE_VALUE;
	srand(0);					//start from same random number
}

CIOCntrl::~CIOCntrl()
{
}


///////////////////////////////////////////////////////////////////////////
//  Stop process thread proceedure and wait for it to finish
//  Called by main program thread to signal data processing thread to stop
//	and then wait for it to stop before returning.
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::StopIO()
{
	if( m_pProcThread != NULL )		//if thread is already running
	{
		PTTCntrl(FALSE);
		m_fProcThreadQuit = TRUE;	// stop it
		::WaitForSingleObject(m_pProcThread->m_hThread,	// wait for it
								INFINITE);
		delete m_pProcThread;
		m_pProcThread = NULL;
	}
}

///////////////////////////////////////////////////////////////////////////
//  Free resources and exit process thread
//   Called by process thread itself to shut itself off if not signaled by
//  main program thread.
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::StopThread()
{
//  if any resources are open, kill em.
	PTTCntrl(FALSE);
	DeleteResources();
	m_pProcThread->m_bAutoDelete = TRUE;	//make thread clean up after
	m_pProcThread = NULL;  				// itself before it dies
	AfxEndThread(3); 
}

///////////////////////////////////////////////////////////////////////////
//   Called to start the I/O thread running
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::StartIO(CWinPSKDoc* pDoc )
{
	InitPerformance();

	m_pDoc = pDoc;
	if( m_pProcThread == NULL)
	{
		m_fProcThreadQuit = FALSE;
		m_pProcThread = AfxBeginThread(
			(AFX_THREADPROC)ThreadLauncher,	//thread function
			(LPVOID)this,					// ptr to this class
			THREAD_PRIORITY_NORMAL, 		//give worker thread priority
			0,								// same stack size
			CREATE_SUSPENDED,				// don't let it start yet
			NULL );							// same security attributes
		m_pProcThread->m_bAutoDelete = FALSE;	// keep thread object around
		m_pProcThread->ResumeThread ();	//let er rip
	}
}

///////////////////////////////////////////////////////////////////////////
//  Main I/O thread processing loop
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::ProcessLoop()
{
	//allocate buffer for signals
	if( (m_pIOBuf = new double[sizeof(double)*BUF_SIZE]) == NULL )
	{
		AfxMessageBox( _T("Input Memory Error") , MB_ICONSTOP );
		StopThread();
	}
//circular delay line for decimate by 2 FIR.
	if( (m_pDec2Que = new double[sizeof(double)*DEC2_LPFIR_LENGTH] ) == NULL)
	{
		AfxMessageBox( _T("Input Memory Error") , MB_ICONSTOP );
		StopThread();
	}

	while( !m_fProcThreadQuit )
	{
		if(m_pDoc->m_TXState != TX_OFF_STATE)
			ProcessXmit();
		else
			ProcessRcv();
	}
	DeleteResources();		//kill all resources
}

///////////////////////////////////////////////////////////////////////////
//   Called when goes to Receive mode
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::ProcessRcv()
{
INT i;
INT result;
double freq,temp;
INT ErrorCode;

//adjust sample rate value
INT	AdjustedSampRate = m_pDoc->m_WFXSettings.nSamplesPerSec  -
							(11025*m_pDoc->m_pSettings->m_ClockAdj)/1000000;

	m_pSoundDev = NULL;
	if( !m_pDoc->m_pSettings->m_DemoMode )	// if not in test mode
	{
		//instantiate a soundcard object for input and start reading.
		m_pSoundDev = new CSound;
		ASSERT( m_pSoundDev);
		ErrorCode = m_pSoundDev->InOpen( &m_pDoc->m_WFXSettings, BUF_SIZE, 0);
		if( ErrorCode != NO_ERROR )
		{
			if( ErrorCode == MMSYSERR_BADDEVICEID )	//if no soundcard
			{
				m_pDoc->m_pSettings->m_DemoMode = TRUE;	//force into test mode
				m_pPSKMod = new CPSKMod;
				ASSERT( m_pPSKMod);
				m_pPSKMod->InitPSKMod(AdjustedSampRate,
								m_pDoc->m_pSettings->m_TxFreq,
								m_pDoc);
			}
			else
			{
				DisplayError( ErrorCode );
				StopThread();
			}
		}
		m_pPSKMod = NULL;
	}
	else	// if testmode, fire up modulation object to create some test sigs.
	{
		m_pSoundDev = NULL;
		m_pPSKMod = new CPSKMod;
		ASSERT( m_pPSKMod);
		m_pPSKMod->InitPSKMod(AdjustedSampRate,
						m_pDoc->m_pSettings->m_TxFreq,
						m_pDoc);
	}
	PTTCntrl(FALSE);		//make sure PTT is off
	m_pPskDet = new CPskDet;	//create detector object
	ASSERT( m_pPskDet);
	InitDec2Filter();
	m_pPskDet->Init( (double)AdjustedSampRate/2.0,BUF_SIZE/2 );
	freq = (double)m_pDoc->m_pSettings->m_RxFreq;
// main processing loop while receiving
	StartDSPPerformance();
	while( !m_fProcThreadQuit && (m_pDoc->m_TXState==TX_OFF_STATE))	//main loop of the output thread
	{
		if( m_pDoc->m_pSettings->m_DemoMode ) //if test mode don't use soundcard
		{	//create our own signal to decode with some AWGNoise
			m_pPSKMod->CalcPSK( m_pIOBuf, BUF_SIZE );
			AddGaussian(m_pIOBuf, BUF_SIZE, m_pDoc->m_NoiseAmplitude);
			StopDSPPerformance();
			if(	m_pDoc->m_FastDemo)
				Sleep(20);
			else
				Sleep(350);		//make about same rate as soundcard
			StartDSPPerformance();
		}
		else
		{
			//Get input data from soundcard object
			if( (result=m_pSoundDev->InRead( m_pIOBuf, BUF_SIZE )) != BUF_SIZE)
			{
				ErrorCode = m_pSoundDev->GetError();
				if( (ErrorCode == SOUNDIN_ERR_OVERFLOW) ||
					(ErrorCode == SOUNDIN_ERR_TIMEOUT) )
				{								//cpu couldn't keep up
												// so try again
					m_pDoc->m_SoundCardReset = TRUE;
					m_pSoundDev->InClose();
					ErrorCode = m_pSoundDev->InOpen( &m_pDoc->m_WFXSettings,
															BUF_SIZE, 0);
				}
				if( ErrorCode != NO_ERRORS)	//must be some other error so abort
				{
					DisplayError( ErrorCode );
					m_pSoundDev->InClose();
					StopThread();
				}
			}
		}
// downsample from 11025 to 5512.5 Hz.
		ProcDec2FIR( m_pIOBuf, m_pIOBuf , BUF_SIZE);	// 2uS per sample
		EnterCriticalSection(&m_pDoc->m_CriticalSection);
// perform receive function on data block
		freq = m_pPskDet->ProcPskDet(	m_pIOBuf, m_pDoc ,	
										m_pDoc->m_pSettings->m_RxFreq,
										m_pDoc->m_pSettings->m_UseAFC);	//5.25uS per sample
// change rx frequency as AFC value changes it
// Change the frequency sort of proportional to the error value(very lame)
		temp = (double)m_pDoc->m_pSettings->m_RxFreq - freq;
		i = (INT)(temp/3.0);
		if( temp > 3.0)
			m_pDoc->m_pSettings->m_RxFreq -= i;
		else 
			if( temp < -3.0 )
				m_pDoc->m_pSettings->m_RxFreq -= i;
			else
				if( temp > 1.1)
					m_pDoc->m_pSettings->m_RxFreq -= 1;
				else 
					if( temp < -1.1)
						m_pDoc->m_pSettings->m_RxFreq += 1;
		LeaveCriticalSection(&m_pDoc->m_CriticalSection);
// Update Display system data array if not busy.
		if( !m_pDoc->m_DispBusy )
		{
			switch( m_pDoc->m_pSettings->m_TabSel )
			{
				case FFT_DISPVIEW:
				case WATERFALL_DISPVIEW:
				case INPUT_DISPVIEW:	//copy into a separate buffer
					for(i=0;i<BUF_SIZE/2;i++)
						m_pDoc->m_pInputData[i] = m_pIOBuf[i];
					break;
				default:
					break;
			}
			SetEvent( m_pDoc->m_EventHandles[EVENT_NEWDATA]);	//tell display
		}
	}
	if(!m_fProcThreadQuit)	//is changing to transmit so close recv resources
	{
		if(m_pPSKMod)
		{
			delete m_pPSKMod;
			m_pPSKMod = NULL;
		}
		if(	m_pSoundDev)
		{
			m_pSoundDev->InClose();
			delete m_pSoundDev;
			m_pSoundDev = NULL;
		}
		if(m_pPskDet)
		{
			delete m_pPskDet;
			m_pPskDet = NULL;
		}
	}
}

///////////////////////////////////////////////////////////////////////////
//   Called when going to Transmit mode
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::ProcessXmit()
{
INT i;
INT result;
INT ErrorCode;
	m_pSoundDev = NULL;
//adjust sample rate value
INT	AdjustedSampRate = m_pDoc->m_WFXSettings.nSamplesPerSec  -
							(11025*m_pDoc->m_pSettings->m_ClockAdj)/1000000;
//instantiate a soundcard object for output and start writing.
	if( !m_pDoc->m_pSettings->m_DemoMode ) //if not in test mode
	{
		m_pSoundDev = new CSound;	//create sound object
		ASSERT( m_pSoundDev);
		ErrorCode = m_pSoundDev->OutOpen( &m_pDoc->m_WFXSettings, BUF_SIZE, 0);
		if( ErrorCode != NO_ERROR )
		{
			DisplayError( ErrorCode );
			StopThread();
		}
		PTTCntrl(TRUE);
	}
	m_pPSKMod = new CPSKMod;	//create modulation object
	ASSERT( m_pPSKMod);
	m_pPSKMod->InitPSKMod(AdjustedSampRate,
						m_pDoc->m_pSettings->m_TxFreq,
						m_pDoc);
	InitDec2Filter();
// loop while transmitting
	StartDSPPerformance();
	while( !m_fProcThreadQuit && (m_pDoc->m_TXState!=TX_OFF_STATE))
	{
		m_pPSKMod->CalcPSK( m_pIOBuf, BUF_SIZE );	//create psk data block
		if( !m_pDoc->m_pSettings->m_DemoMode ) //if test mode don't use soundcard
		{
			if( (result=m_pSoundDev->OutWrite( m_pIOBuf, BUF_SIZE )) != BUF_SIZE)
			{
				ErrorCode = m_pSoundDev->GetError();
				if( (ErrorCode == SOUNDOUT_ERR_UNDERFLOW) ||
					(ErrorCode == SOUNDOUT_ERR_TIMEOUT) )
				{								//cpu couldn't keep up
												// so try again
					m_pDoc->m_SoundCardReset = TRUE;
					m_pSoundDev->OutClose();
					ErrorCode = m_pSoundDev->OutOpen( &m_pDoc->m_WFXSettings, BUF_SIZE, 0);
				}
				if( ErrorCode != NO_ERRORS)	//must be some other soundcard error
				{
					DisplayError( m_ErrorCode );
					m_pSoundDev->OutClose();
					StopThread();
				}
			}
		}
		else
		{
			StopDSPPerformance();
			if(	m_pDoc->m_FastDemo)
				Sleep(10);
			else
				Sleep(350);
			StartDSPPerformance();
		}
// Update Display system data array if not busy.
// downsample for display system
		ProcDec2FIR( m_pIOBuf, m_pIOBuf , BUF_SIZE);
		if( !m_pDoc->m_DispBusy )
		{
			switch( m_pDoc->m_pSettings->m_TabSel )
			{
				case FFT_DISPVIEW:
				case WATERFALL_DISPVIEW:
				case INPUT_DISPVIEW:
					for(i=0; i<BUF_SIZE/2; i++)
						m_pDoc->m_pInputData[i] = m_pIOBuf[i]/3.1620;	//pure sinewave=-10dB
					SetEvent( m_pDoc->m_EventHandles[EVENT_NEWDATA]);
					break;
				default:
					break;
			}
		}
	}
	if(!m_fProcThreadQuit)	//is changing to rcv so close xmit resources
	{
		if(	m_pSoundDev)
		{
			m_pSoundDev->OutClose();
			delete m_pSoundDev;
			m_pSoundDev = NULL;
		}
		if(m_pPSKMod)
		{
			delete m_pPSKMod;
			m_pPSKMod = NULL;
		}
	}

}

/////////////////////////////////////////////////////////////////
//	Decimate by 2 FIR filter on 'BlockSize' samples.
//	pIn == pointer to input array of double's (can be same buffer as pOut )
//	pOut == pointer to output array of double's
//	Blocksize == number of samples to process
/////////////////////////////////////////////////////////////////
void CIOCntrl::ProcDec2FIR( double* pIn, double* pOut , INT BlockSize)
{
INT i,j;
double acc;				
const double* Kptr;
double* Firptr;
double* Qptr;
double* Inptr;
	Inptr = m_pDec2InPtr;		//use automatic copies of member variables
	Qptr = m_pDec2Que;			// for better speed.
	j = 0;
	for( i = 0; i<BlockSize; i++ )	// put new samples into Queue
	{
		if( --Inptr < Qptr )		//deal with wraparound
			Inptr = Qptr+DEC2_LPFIR_LENGTH-1;
		*Inptr = pIn[i];
		if( i&1 )		//calculate MAC's every other time for decimation by 2
		{
			acc = 0.0;
			Firptr = Inptr;
			Kptr = Dec2LPCoef;
			while( Kptr < (Dec2LPCoef + DEC2_LPFIR_LENGTH) )	//do the MAC's
			{
				acc += ( (*Firptr++)*(*Kptr++) );
				if( Firptr >= Qptr+DEC2_LPFIR_LENGTH )	//deal with wraparound
					Firptr = Qptr;
			}
			pOut[j++] = acc;		//save output sample
		}
	}
	m_pDec2InPtr = Inptr;		// save position in circular delay line
}

/////////////////////////////////////////////////////////////////
//	Initialize Decimate by 2 FIR filter buffers.
/////////////////////////////////////////////////////////////////
void CIOCntrl::InitDec2Filter()
{
	m_pDec2InPtr = m_pDec2Que;
	for( INT i=0; i<DEC2_LPFIR_LENGTH; i++)
	{
		m_pDec2Que[i] = 0.0;	// fill delay buffer with zero
	}

}

///////////////////////////////////////////////////////////////////////////
//   Called to delete any allocated resources in this class
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::DeleteResources()
{
	if(	m_pSoundDev)
	{
		m_pSoundDev->InClose();
		m_pSoundDev->OutClose();
		delete m_pSoundDev;
		m_pSoundDev = NULL;
	}
	if(m_pIOBuf)
	{
		delete m_pIOBuf;	
		m_pIOBuf = NULL;
	}
	if(m_pDec2Que)
	{
		delete m_pDec2Que;
		m_pDec2Que = NULL;
	}
	if(m_pPSKMod)
	{
		delete m_pPSKMod;	
		m_pPSKMod = NULL;
	}
	if(m_pPskDet)
	{
		delete m_pPskDet;
		m_pPskDet = NULL;
	}
}
//////////////////////////////////////////////////////////////////////////
// Simple minded error message display for various sound/wave errors
// Should be put in string table.
//////////////////////////////////////////////////////////////////////////
void CIOCntrl::DisplayError( INT ErrorCode )
{
CString Msg;
	switch( ErrorCode )
	{
		case NO_ERRORS:
			Msg = _T("");
			break;
		case MEMORY_ERROR:
			Msg = _T("Memory Allocation error");
			break;
		case WAVIN_ERR_OPEN:
			Msg = _T("Can't open wave file for input");
			break;
		case WAVIN_ERR_NOTWAVE:
			Msg = _T("File is not a RIFF wave type");
			break;
		case WAVIN_ERR_INVALID:
			Msg = _T("Invalid wave file");
			break;
		case WAVIN_ERR_NODATA:
			Msg = _T("No data in file");
			break;
		case WAVIN_ERR_NOTSUPPORTED:
			Msg = _T("Not a supported data type");
			break;
		case WAVIN_ERR_READING:
			Msg = _T("Error reading data from file");
			break;
		case WAVIN_ERR_NOTOPEN:
			Msg = _T("Tried to read and file is not open");
			break;
		case WAVOUT_ERR_OPEN:
			Msg = _T("Can't open wave file for input");
			break;
		case WAVOUT_ERR_WRITING:
			Msg = _T("Error writing to wave file");
			break;
		case WAVOUT_ERR_NOTOPEN:
			Msg = _T("Tried to write and file is not open");
			break;
		case SOUNDIN_ERR_NOTOPEN:
			Msg = _T("Tried to read and soundcard is not open");
			break;
		case SOUNDIN_ERR_OVERFLOW:
			Msg = _T("Input buffers overflowed");
			break;
		case SOUNDIN_ERR_TIMEOUT:
			Msg = _T("Timed out waiting for input buffers");
			break;
		case SOUNDOUT_ERR_NOTOPEN:
			Msg = _T("Tried to write and soundcard is not open");
			break;
		case SOUNDOUT_ERR_UNDERFLOW:
			Msg = _T("Output buffers Underflowed");
			break;
		case SOUNDOUT_ERR_TIMEOUT:
			Msg = _T("Timed out waiting for output buffers");
			break;
//
		case MMSYSERR_BADDEVICEID:
			Msg = _T("Bad Device ID\nIs Soundcard Present?");
			break;
		case MMSYSERR_NOTENABLED:
			Msg = _T("Driver failed enable");
			break;
		case MMSYSERR_ALLOCATED:
			Msg = _T("Device already allocated");
			break;
		case MMSYSERR_INVALHANDLE:
			Msg = _T("Device handle is invalid");
			break;
		case MMSYSERR_NODRIVER:
			Msg = _T("No device driver present");
			break;
		case MMSYSERR_NOMEM:
			Msg = _T("mmsystem memory allocation error");
			break;
		case MMSYSERR_NOTSUPPORTED:
			Msg = _T("Function isn't supported");
			break;
		case MMSYSERR_BADERRNUM:
			Msg = _T("Error value out of range");
			break;
		case MMSYSERR_INVALFLAG: 
			Msg = _T("Invalid flag passed");
			break;
		case MMSYSERR_INVALPARAM: 
			Msg = _T("Invalid parameter passed");
			break;
		case MMSYSERR_HANDLEBUSY:
			Msg = _T("Handle being used");
			break;
		default:
			Msg = _T("??? Maybe Input being used at\n a different sample rate");
			break;
	}
	if( Msg != _T("") )
		AfxMessageBox( Msg );
}

//////////////////////////////////////////////////////////////////
//  Controls COM port signals DTR and RTS for Xmitter on/off control
//////////////////////////////////////////////////////////////////
void CIOCntrl::PTTCntrl(BOOL Ptton)
{
char szPort[10];
DCB dcb;
	if( Ptton  )
	{
		if( m_pDoc->m_pSettings->m_PTTPort == 0 )
			return;
		if( m_hComPort == INVALID_HANDLE_VALUE )
		{
			wsprintf( szPort, "COM%d", m_pDoc->m_pSettings->m_PTTPort );
			m_hComPort = CreateFile( szPort, GENERIC_READ | GENERIC_WRITE, 0, NULL,
								OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
			if( m_hComPort == INVALID_HANDLE_VALUE )
				return;
			dcb.DCBlength = sizeof( DCB );
			GetCommState( m_hComPort, &dcb );
			dcb.fDtrControl = DTR_CONTROL_ENABLE;
			dcb.fRtsControl = RTS_CONTROL_ENABLE;
			SetCommState( m_hComPort, &dcb );
		}
	}
	else
	{
		if( m_hComPort != INVALID_HANDLE_VALUE )
		{
			dcb.DCBlength = sizeof( DCB );
			GetCommState( m_hComPort, &dcb );
			dcb.fDtrControl = DTR_CONTROL_DISABLE;
			dcb.fRtsControl = RTS_CONTROL_DISABLE;
			SetCommState( m_hComPort, &dcb );
			CloseHandle( m_hComPort );
			m_hComPort = INVALID_HANDLE_VALUE;
		}
	}
}

//////////////////////////////////////////////////////////////////
// Adds n gaussian random doubles with 0 mean and std_dev = RMS = std
//    to the specified buffer
//////////////////////////////////////////////////////////////////
void CIOCntrl::AddGaussian(double * pData, INT n, double std)
{
INT i = 0;
double rad;
double r;
double u1;
double u2;
	while( i<n )
	{
// Generate two uniform random numbers between -1 and +1
// that are inside the unit circle
		do {
			u1 = 1.0 - 2.0 * (double)rand()/(double)RAND_MAX ;
			u2 = 1.0 - 2.0 * (double)rand()/(double)RAND_MAX ;
			r = u1*u1 + u2*u2;
		} while(r >= 1.0 || r == 0.0);
		rad = sqrt(-2.0*log(r)/r);
		pData[i++] += (std*u1*rad);
		pData[i++] += (std*u2*rad);
	}
}


