// PlotData.cpp: implementation of the CPlotData class.
//
//

#include "stdafx.h"
#include "winpsk.h"
#include "Settings.h"
#include "WinPSKView.h"
#include "PlotData.h"
#include "PlotTables.h"
#include "Perform.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// Non C++ thread routine used for launching the drawing thread
static UINT DrawThreadLauncher(CPlotData* pCPlotData )
{
	pCPlotData->DrawLoop( );
	return(6);
}


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CPlotData::CPlotData()
{
	m_pDoc = NULL;
	m_pDrawThread = NULL;
	m_StopDraw = FALSE;
	m_pFft = NULL;
	m_DispHead = 0;
	m_DispTail = 0;
	m_BufToggle = FALSE;
	m_NeedToClrBufs = FALSE;
	m_LastOvrLoad = FALSE;
//allocate memory space
	m_pDispBuf = new INT[MAX_DISPBUF+10];
	ASSERT( m_pDispBuf );
	ClearBuffers();
	m_TranslateTbl = new INT[MAX_DISPBUF+10];
	ASSERT( m_TranslateTbl );
//allocate FFT object
	m_pFft = new Cfft;
	m_CursPos = 0;
	m_pPenR = new CPen;		// create some colored pens for drawing
	m_pPenG = new CPen;
	m_pPenY = new CPen;
	m_pPenW = new CPen;
	m_pPenC = new CPen;
	m_pPenB = new CPen;
	m_pPenR->CreatePen( PS_SOLID,0, RGB(255,0,0) );
	m_pPenG->CreatePen( PS_SOLID,0, RGB(0,255,0) );
	m_pPenY->CreatePen( PS_DOT,0, RGB(255,255,0) );
	m_pPenW->CreatePen( PS_SOLID,0, RGB(128,128,128) );
	m_pPenC->CreatePen( PS_SOLID,0, RGB(0,255,255) );
	m_pPenB->CreatePen( PS_SOLID,0, RGB(0,0,255) );
}

CPlotData::~CPlotData()
{
	m_pPenR->DeleteObject(); delete m_pPenR;
	m_pPenG->DeleteObject(); delete m_pPenG;
	m_pPenY->DeleteObject(); delete m_pPenY;
	m_pPenW->DeleteObject(); delete m_pPenW;
	m_pPenC->DeleteObject(); delete m_pPenC;
	m_pPenB->DeleteObject(); delete m_pPenB;
	if(m_pDispBuf)
	{
		delete m_pDispBuf;
		m_pDispBuf = NULL;
	}
	if(m_pFft)
	{
		delete m_pFft;
		m_pFft = NULL;
	}	
	if(m_TranslateTbl)
	{
		delete m_TranslateTbl;
		m_TranslateTbl = NULL;
	}	
// Delete two ping pong bitmap buffers for waterfall display
	if( m_bmWaterA.m_hObject != NULL )	// delete old bitmap object
		m_bmWaterA.DeleteObject();
	if( m_bmWaterB.m_hObject != NULL )	// delete old bitmap object
		m_bmWaterB.DeleteObject();
}

//////////////////////////////////////////////////////////////////////
// Wait for new data to display in worker drawing thread 
//////////////////////////////////////////////////////////////////////
void CPlotData::DrawLoop( )
{
DWORD event;
	while(!m_StopDraw)
	{
// wait for the event(new data is ready or plot cursor update)
		m_pDoc->m_DispBusy = FALSE;
		event = ::WaitForMultipleObjects( 2,m_pDoc->m_EventHandles,FALSE,INFINITE );
		m_pDoc->m_DispBusy = TRUE;
		if(!m_StopDraw)
		{
			StartDISPPerformance();
			DrawPlot( event);
			StopDISPPerformance();
		}
	}
}

//////////////////////////////////////////////////////////////////////
//       Starts the data plotting thread
//////////////////////////////////////////////////////////////////////
void CPlotData::StartDrawThread()
{
	m_pDrawThread = AfxBeginThread((AFX_THREADPROC)DrawThreadLauncher,
				(LPVOID)this,					// handle to this view
				THREAD_PRIORITY_BELOW_NORMAL,	//give worker thread lower priority
				0,								// same stack size
				CREATE_SUSPENDED,				// don't let it start yet
				NULL );							// same security attributes
	m_pDrawThread->m_bAutoDelete = FALSE;		// keep thread object around
	m_StopDraw = FALSE;							// let the plotting thread begin
	m_pDrawThread->ResumeThread ();
}

//////////////////////////////////////////////////////////////////////
//       Stops the data plotting thread
//////////////////////////////////////////////////////////////////////
void CPlotData::StopDrawThread()
{
	if (m_pDrawThread)		//if drawing thread is already running
	{
		m_StopDraw = TRUE;	// signal to stop it
		m_pDrawThread->ResumeThread ();	//in case thread has been suspended
		SetEvent( m_pDoc->m_EventHandles[EVENT_SCRNUPDATE]);
		::WaitForSingleObject(m_pDrawThread->m_hThread, INFINITE);	// wait as long as necessary
		delete m_pDrawThread;
		m_pDrawThread = NULL;
	}
}

//////////////////////////////////////////////////////////////////////
//  Called from the Windows view class when the window has changed
//      and needs to update the screen coordinates.
//////////////////////////////////////////////////////////////////////
void CPlotData::OnDrawPlot(HWND hWnd, CFormView* pView, CWinPSKDoc* pDoc)
{
CRect rScrn;
	m_pDoc = pDoc;
	m_hViewWnd = hWnd;
	StopDrawThread();	//stop the current drawing thread
						//calculate new plot area coordinates
// Get pixel size of plot area bitmap
	ASSERT_VALID( pView );
	CWnd* pWnd = pView->GetDlgItem( IDC_PLOTAREA );
	ASSERT_VALID( pWnd);
	pWnd->GetWindowRect( m_rPlot );
	pView->GetWindowRect(rScrn);
	m_rPlot.top -= (rScrn.top+1);
	m_rPlot.left -= (rScrn.left+1);
	m_rPlot.bottom = m_rPlot.top+m_rPlotsize.bottom;
	m_rPlot.right = m_rPlot.left+m_rPlotsize.right;
// Create two ping pong bitmap buffers for waterfall display
	if( m_bmWaterA.m_hObject != NULL )	// delete old bitmap object
		m_bmWaterA.DeleteObject();
	if( m_bmWaterB.m_hObject != NULL )	// delete old bitmap object
		m_bmWaterB.DeleteObject();
//create a 64 line waterfall ping pong bitmap with current screen DC settings
CClientDC ClientDC(CWnd::FromHandle(m_hViewWnd) );
	m_bmWaterA.CreateCompatibleBitmap( &ClientDC, m_rPlot.Width(), m_rPlotsize.bottom );
	m_bmWaterB.CreateCompatibleBitmap( &ClientDC, m_rPlot.Width(), m_rPlotsize.bottom );
	m_NeedToClrBufs = TRUE;		// so waterfal buffers can get cleared
//
	StartDrawThread();		// restart screen updates
	SetEvent( m_pDoc->m_EventHandles[EVENT_SCRNUPDATE]);
}


//////////////////////////////////////////////////////////////////////
//  Called to clear plot surface, draw cursor, and call specific plot
//         routine.
//////////////////////////////////////////////////////////////////////
void CPlotData::DrawPlot(DWORD event)
{
// erases plot surface
// plots data points
BOOL newdata;
INT i;
CDC MemDC;
CPen* pOldPen;
CClientDC ClientDC(CWnd::FromHandle(m_hViewWnd) );
	m_InputOverload = FALSE;
	newdata = (event == (WAIT_OBJECT_0 + EVENT_NEWDATA) );
	if( m_bmPlot.m_hObject != NULL )	// delete old bitmap object
		m_bmPlot.DeleteObject();
//create a new bitmap with current screen DC settings
	m_bmPlot.CreateCompatibleBitmap( &ClientDC, m_rPlot.Width(),
					 m_rPlot.Height() );
//create memory DC pointer to the plot bitmap
	MemDC.CreateCompatibleDC( &ClientDC );
CBitmap* pbmOld = MemDC.SelectObject( &m_bmPlot );

// set bitmap background to black using slightly faster ExtTextOut function
	MemDC.SetBkColor( RGB(0, 0, 0) );
	MemDC.ExtTextOut( 0, 0,ETO_OPAQUE, &m_rPlotsize,NULL,0,NULL);

	pOldPen = MemDC.SelectObject(m_pPenY);		//save original pen.
// draw selected data plot
	switch( m_pDoc->m_pSettings->m_TabSel )
	{
		case FFT_DISPVIEW:
			// Draw vertical yellow dotted plot cursor line
			MemDC.MoveTo( m_CursPos, 0);
			MemDC.LineTo( m_CursPos, m_rPlotsize.bottom );
			if( m_pDoc->m_pSettings->m_fSlowAve)
				m_FFTAveVal = 6;
			else
				m_FFTAveVal = 2;
			MemDC.SelectObject(m_pPenW);
			for( i=10; i<100; i+=10 )	//draw 10dB tic marks
			{
				MemDC.MoveTo( 0, (m_rPlotsize.bottom*i)/100 );
				MemDC.LineTo( m_rPlotsize.right, (m_rPlotsize.bottom*i)/100 );
			}
			MemDC.SelectObject(m_pPenG);
			PlotFFTData( &MemDC, newdata );
			MemDC.SelectObject(m_pPenB);
			PlotOverlayVectData( &MemDC );
			break;
		case WATERFALL_DISPVIEW:
			// Draw vertical yellow dotted plot cursor line
			MemDC.MoveTo( m_CursPos, 0);
			MemDC.LineTo( m_CursPos, m_rPlotsize.bottom );
			if( m_pDoc->m_pSettings->m_fSlowAve)
				m_FFTAveVal = 4;
			else
				m_FFTAveVal = 1;
			PlotFFTWater( &MemDC, newdata );
			MemDC.SelectObject(m_pPenB);
			PlotOverlayVectData( &MemDC );
			break;
		case INPUT_DISPVIEW:
			MemDC.SelectObject(m_pPenG);
			PlotInputData( &MemDC );
			break;
		case SYNC_DISPVIEW:
			PlotSyncData( &MemDC);
			break;
		default:
			break;
	}
	if(m_InputOverload && m_LastOvrLoad) //need two in a row
	{
		MemDC.SetTextColor( RGB(255,0,0) );
		MemDC.TextOut( m_rPlotsize.right/4, m_rPlotsize.bottom/8,
			_T("Reduce Audio Level"), 18 );
	}
	m_LastOvrLoad = m_InputOverload;
	if( m_pDoc->m_SoundCardReset)
	{
		MemDC.SetTextColor( RGB(255,0,0) );
		MemDC.TextOut( m_rPlotsize.right/4, m_rPlotsize.bottom/8,
			_T("CPU Too Slow"), 12 );
		m_pDoc->m_SoundCardReset = FALSE;
	}
// Copy memory bitmap into screen device context for flicker free update
	ClientDC.BitBlt( m_rPlot.left, m_rPlot.top,
					m_rPlotsize.right,
					m_rPlotsize.bottom, &MemDC,
					0,0,SRCCOPY);
	MemDC.SelectObject(pOldPen);	//deselect pens and bitmap
	MemDC.SelectObject( pbmOld );	// from memory DC
}



//////////////////////////////////////////////////////////////////////
//		Perform FFT and plot two dimensional results
//////////////////////////////////////////////////////////////////////
void CPlotData::PlotFFTData( CDC* pDC, BOOL newdata)
{
INT x,y,k1;
DOUBLE k2;
	k1 = m_rPlotsize.bottom;		//y range
	k2 = 0.1*(DOUBLE)k1;			//y scale value
	if(newdata)
	{
		m_InputOverload = m_pFft->CalcFFT(m_pDoc->m_pInputData, m_XMinPos,
						m_XMaxPos,k2 , m_FFTAveVal, m_pDispBuf);
	}
	pDC->MoveTo( 0, k1-m_pDispBuf[m_XMinPos]);//move to first position
	for( x=m_XMinPos; x<m_XMaxPos; x++ )	//~11mSecs on 266PII
	{
		y=k1-m_pDispBuf[x];
		pDC->LineTo( m_TranslateTbl[x], y );		//draw to the next pt.
	}
}


//////////////////////////////////////////////////////////////////////
//		Perform FFT and plot waterfall
//////////////////////////////////////////////////////////////////////
void CPlotData::PlotFFTWater( CDC* pDC, BOOL newdata)
{
INT i,x,k1,k2,k3,k4;
COLORREF cr;
CDC MemDCA;		// memory DCs for the above bitmaps
CDC MemDCB;
	k4 = m_rPlotsize.bottom;
//create memory DC pointers for the waterfall bitmap buffers
	MemDCA.CreateCompatibleDC( pDC );
	MemDCB.CreateCompatibleDC( pDC );
CBitmap* pbmOldA = MemDCA.SelectObject( &m_bmWaterA );
CBitmap* pbmOldB = MemDCB.SelectObject( &m_bmWaterB );
	if(m_NeedToClrBufs)
	{
		MemDCA.BitBlt( 0, 0, m_rPlotsize.right,k4, NULL, 0,0,BLACKNESS);
		MemDCB.BitBlt( 0, 0, m_rPlotsize.right,k4, NULL, 0,0,BLACKNESS);
		m_NeedToClrBufs = FALSE;
	}
	k3 = k4/4; //y starting point
	if(newdata)
	{
		m_InputOverload = m_pFft->CalcFFT(m_pDoc->m_pInputData,MIN_DISPBUF,
						MAX_DISPBUF-1, 25.5, m_FFTAveVal, m_pDispBuf);
	}
	k2 = (m_XMaxPos-m_XMinPos);
//
	if(m_BufToggle) //use A buffer
	{
		// Copy lines from buf B to A shifted down 2 lines
		MemDCA.BitBlt( 0, 2, m_rPlotsize.right,k4-k3-2, &MemDCB, 0,0,SRCCOPY);
		// Draw top two new lines int to buf A
		for( x=0; x<m_rPlotsize.right; x++)
		{
			k1 = (x*k2)/m_rPlotsize.right + m_XMinPos;
			i = m_pDispBuf[k1];
			cr = ColorTbl[i];		//RGB( i, i, i);
			MemDCA.SetPixel( x , 0, cr );
			MemDCA.SetPixel( x , 1, cr );
		}
		// copy buf A into main plot bitmap
		pDC->BitBlt( 0, k3,	m_rPlotsize.right,k4-k3, &MemDCA, 0,0,SRCCOPY);
	}
	else			// use B buffer
	{
		// Copy lines from buf A to B shifted down 2 lines
		MemDCB.BitBlt( 0, 2, m_rPlotsize.right,k4-k3-2, &MemDCA, 0,0,SRCCOPY);
		// Draw top two new lines int to buf B
		for( x=0; x<m_rPlotsize.right; x++)
		{
			k1 = (x*k2)/m_rPlotsize.right + m_XMinPos;
			i = m_pDispBuf[k1];
			cr = ColorTbl[i];		//RGB( i, i, i);
			MemDCB.SetPixel( x , 0, cr );
			MemDCB.SetPixel( x , 1, cr );
		}
		// copy buf B into main plot bitmap
		pDC->BitBlt( 0, k3,	m_rPlotsize.right,k4-k3, &MemDCB, 0,0,SRCCOPY);
	}
	MemDCA.SelectObject( pbmOldA );
	MemDCB.SelectObject( pbmOldB );
	m_BufToggle = !m_BufToggle;
}


//////////////////////////////////////////////////////////////////////
// Plot Baseband Data as vector plot on top of another graph
//////////////////////////////////////////////////////////////////////
void CPlotData::PlotOverlayVectData( CDC* pDC )
{
INT x,y,xc,yc,x1,y1;
double mag;
double tmpx;
double tmpy;
	yc = m_rPlotsize.bottom/8;	// y center
	xc = m_rPlotsize.right/8;	// x center
	pDC->Arc( xc-yc, 0, xc+yc, m_rPlotsize.bottom/4, xc ,0, xc,0);	//draw circle
	for( x=196,y=0; x<10*196; y+=196, x+=196 )// 196 = (612.5/31.25)*10
	{
		x1 = x/10;
		y1 = y/10;
		tmpx = m_pDoc->m_pIQDispData[y1].x * m_pDoc->m_pIQDispData[x1].x +
				m_pDoc->m_pIQDispData[y1].y * m_pDoc->m_pIQDispData[x1].y;
		tmpy = m_pDoc->m_pIQDispData[y1].y * m_pDoc->m_pIQDispData[x1].x -
				m_pDoc->m_pIQDispData[y1].x * m_pDoc->m_pIQDispData[x1].y;
		pDC->MoveTo( xc,  yc );
		mag = sqrt(tmpx * tmpx + tmpy * tmpy);
		pDC->SelectObject(m_pPenC);
		pDC->LineTo( xc+(INT)(yc*tmpy/mag), yc-(INT)(yc*tmpx/mag) );
	}
}

//////////////////////////////////////////////////////////////////////
// Plot Sync position distribution
//////////////////////////////////////////////////////////////////////
void CPlotData::PlotSyncData( CDC* pDC )
{
INT x,y,i,k1,k2;
INT ys;
	y = m_rPlotsize.bottom;
	k1 = 13000/y;
	k2 = m_rPlotsize.right/20;
	pDC->SelectObject(m_pPenG);
	for( x=k2/2,i=0; i<20; i++,x+=k2)
	{
		ys = m_pDoc->m_SyncHist[i]/k1;
		m_pDoc->m_SyncHist[i] = 0;
		pDC->MoveTo( x, y );
		pDC->LineTo( x, y-ys );
	}
}

//////////////////////////////////////////////////////////////////////
// Plots the input data directly in the time domain
//////////////////////////////////////////////////////////////////////
void CPlotData::PlotInputData( CDC* pDC)
{
INT x,y,z,k1;
BOOL overload = FALSE;
	k1 = m_rPlotsize.bottom/2;		//center
	pDC->MoveTo( 0, k1);			//move to first position
	for( x=0; x<m_rPlotsize.right; x++ )
	{
		z = (INT)m_pDoc->m_pInputData[x];
		if(!overload)
		{
			if( (z > 16384) || (z < -16384) )
			{
				pDC->SelectObject(m_pPenR);
				overload = TRUE;
			}
		}
		y = k1 + (k1*z)/16384;
		pDC->LineTo( x, y );	//draw to the next pt.
	}
}


//////////////////////////////////////////////////////////////////////
// Sets the plot zoom factor and calculates all the new parameters.
//////////////////////////////////////////////////////////////////////
void CPlotData::SetZoom( INT zoom, BOOL Cent)
{
INT xdelta;
CString Msg;
	m_Zoom = zoom;
	xdelta = (MAX_DISPBUF-MIN_DISPBUF+1)/zoom;
	if( !Cent )
	{
		if( (m_XCurPos <= m_XMinPos) || (m_XCurPos >= m_XMaxPos) )
			Cent = TRUE;
	}
	if( Cent )
	{
		// Try to center the cursor
		m_XMinPos = m_XCurPos - xdelta/2;
		m_XMaxPos = m_XCurPos + xdelta/2;
		if( m_XMinPos<MIN_DISPBUF )
		{
			m_XMinPos = MIN_DISPBUF;
			m_XMaxPos = xdelta + MIN_DISPBUF;
		}
		else
		{
			if( m_XMaxPos > MAX_DISPBUF )
			{
				m_XMinPos = MAX_DISPBUF-xdelta;
				if( m_XMinPos<MIN_DISPBUF )
					m_XMinPos = MIN_DISPBUF;
				m_XMaxPos = MAX_DISPBUF;
			}
		}
	}
	m_CursPos = m_rPlotsize.right*( m_XCurPos - m_XMinPos)
					/(m_XMaxPos - m_XMinPos); 
// The translate table is used to map DispBuf[] index's into
//		Screen plot coordinates.
	for(INT i=m_XMinPos; i<m_XMaxPos; i++)
	{
		m_TranslateTbl[i] = ( (i-m_XMinPos)*m_rPlotsize.right )
									/ (m_XMaxPos - m_XMinPos);
	}
}



//////////////////////////////////////////////////////////////////////
// Called when a user clicks on the plot window to place the plot cursor.
// Sets the plot cursor to the specified screen position.  The DispBuf[]
//   index is then calculated.  The cursor on the screen is then updated.
//////////////////////////////////////////////////////////////////////
void CPlotData::SetPlotCursorPos( INT pos )
{
	m_CursPos = pos - m_rPlot.left;
	m_XCurPos = (m_CursPos*(m_XMaxPos - m_XMinPos)
					/m_rPlotsize.right) + m_XMinPos;
	SetEvent( m_pDoc->m_EventHandles[EVENT_SCRNUPDATE]);
}

//////////////////////////////////////////////////////////////////////
// Called to clear the display buffers
//////////////////////////////////////////////////////////////////////
void CPlotData::ClearBuffers()
{
	for(INT i=0; i<MAX_DISPBUF; i++)
		m_pDispBuf[i] = 0;
	m_NeedToClrBufs = TRUE;		// so waterfal buffers can get cleared
}

