//////////////////////////////////////////////////////////////////////
// PskDet.cpp: implementation of the CPskDet class.
//////////////////////////////////////////////////////////////////////
//

#include "stdafx.h"
#include "WinPSK.h"
#include "WinPSKDoc.h"
#include "PskDet.h"
#include "PskDetTbl.h"
#include "PskTables.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// define some constants
#define PI2 ( 8.0 * atan(1.0) )		// 2 Pi
#define PHZ_180_MIN	(PI2/8.0)		//Pi/4
#define PHZ_180_MAX	(3.0*PI2/8.0)	//3Pi/4
#define PHZ_0_MIN	(5.0*PI2/8.0)	//5Pi/4
#define PHZ_0_MAX	(7.0*PI2/8.0)	//7Pi/4
#define Ts (.032+.000000)			// Ts == symbol period(correct for +/- .23%error)
//#define Ts (.032+.000072)		//to simulate off freq soundcards

//define the PSK31 symbols
#define SYM_NOCHANGE 0	//Stay the same phase
#define SYM_P90 1		//Plus 90  deg
#define SYM_P180 2		//Plus 180 deg
#define SYM_M90 3		//Minus 90 deg

// phase wraparound correction table for viterbi decoder
const double AngleTbl1[4] = { 3.0*PI2/4.0, 0.0, PI2/4.0, PI2/2.0};
const double AngleTbl2[4] = { 3.0*PI2/4.0, PI2, PI2/4.0, PI2/2.0 };


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPskDet::CPskDet()
{
	m_pQue1 = NULL;
	m_pQue2 = NULL;
	m_pQue3 = NULL;
	m_VaricodeDecTbl = NULL;
}

CPskDet::~CPskDet()
{
	if(m_pQue1)
		delete m_pQue1;
	if(m_pQue2)
		delete m_pQue2;
	if(m_pQue3)
		delete m_pQue3;
	if(m_VaricodeDecTbl)
		delete m_VaricodeDecTbl;
}

/////////////////////////////////////////////////////////////////
//       Initialize PskDet buffers and pointers
/////////////////////////////////////////////////////////////////
BOOL CPskDet::Init( double Fs, INT BlockSize )
{
WORD wTemp;
//circular delay lines.(data stays put and the pointers move)
	if( (m_pQue1 = new _complex[sizeof(_complex)*DEC3_LPFIR_LENGTH] ) == NULL)
		return FALSE;
	if( (m_pQue2 = new _complex[sizeof(_complex)*DEC3_LPFIR_LENGTH] ) == NULL)
		return FALSE;
	if( (m_pQue3 = new _complex[sizeof(_complex)*BITFIR_LENGTH] ) == NULL)
		return FALSE;
	if( (	m_VaricodeDecTbl = new BYTE[sizeof(BYTE)*2048] )== NULL)
		return FALSE;
	for(INT i=0; i<DEC3_LPFIR_LENGTH; i++)
	{
		m_pQue1[i].x = 0.0;	// fill delay buffer with zero
		m_pQue1[i].y = 0.0;
	}
	for(i=0; i<DEC3_LPFIR_LENGTH; i++)
	{
		m_pQue2[i].x = 0.0;	// fill delay buffer with zero
		m_pQue2[i].y = 0.0;
	}
	for( i=0; i<BITFIR_LENGTH; i++)
	{
		m_pQue3[i].x = 0.0;	// fill delay buffer with zero
		m_pQue3[i].y = 0.0;
	}
	for( WORD j=0; j<2048; j++)		//init inverse varicode lookup decoder table
	{
		m_VaricodeDecTbl[j] = 0;
		for( i=0; i<256;i++)
		{
			wTemp = VARICODE_TABLE[i];
			wTemp >>= 4;
			while( !(wTemp&1) )
				wTemp >>= 1;
			wTemp >>= 1;
			if( wTemp == j)
				m_VaricodeDecTbl[j] = (BYTE)i;
		}
	}
	m_pInPtr1 = m_pQue1;		//initialize que ptrs
	m_pInPtr2 = m_pQue2;
	m_pInPtr3 = m_pQue3;
	m_Fs = Fs;					//sample rate
	m_BlockSize = BlockSize;	//size data input buffer
	m_phzinc = 1000.0*PI2/m_Fs;	
	m_BitPhaseInc = 9.0/m_Fs;		//bit oversampling period
	m_SignalLevel = 1.0;
	m_BitPhasePos = 0.0;
	m_BitAcc = 0;
	m_LastBitZero = FALSE;
	m_SampCnt = 0;
	m_OnCount = 0;
	m_OffCount = 0;
	for(i=0; i<21; i++)
	{
		m_SyncAve[i] = 0.0;					// initialize the array
		ViterbiDecode( 3.0*PI2/4.0, FALSE);	// init the Viterbi decoder
	}
	for(i=0; i<16; i++)
	{
		m_SurvivorStates[i].Pathdistance = 0.0;
		m_SurvivorStates[i].BitEstimates = 0;
	}
// Init a bunch of "static" variables used in various member fuctions
	m_AGCave = 0.0;
	m_FreqError = 0.0;
	m_VcoPhz = 0.0;
	m_LastFreq = 1000;
	m_PkPos = 0;
	m_NewPkPos = 5;
	m_BitPos = 0;
	m_I3 = 0.0;
	m_I2 = 0.0;
	m_I1 = 0.0;
	m_I0 = 0.0;
	m_Q3 = 0.0;
	m_Q2 = 0.0;
	m_Q1 = 0.0;
	m_Q0 = 0.0;
	m_DevAve = .4;
	m_z1.x = 0.0; m_z1.y = 0.0;
	m_z2.x = 0.0; m_z2.y = 0.0;
	m_FerAve = 0.0;
	m_QFreqError = 0.0;
	m_LastPkPos = 0;
	m_ClkErrCounter = 0;
	m_ClkErrTimer = 0;
	m_ClkError = 0;
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Main routine called to process the next block of data 'pIn'.
//  The center frequency is specified by 'Freq'.
//  The function returns the new AFC'd frequency.
//  'UseAFC' specifies whether to automatically lock to the frequency.
//  30mSec for BPSK, 38mSec for QPSK( 133MHz Pentium )
//////////////////////////////////////////////////////////////////////
double CPskDet::ProcPskDet( double* pIn, CWinPSKDoc* pDoc , INT Freq,  BOOL UseAFC)
{
double freqerror = m_FreqError;
double vcophz = m_VcoPhz;
INT i,j;
_complex acc;
const double* Kptr;
_complex* Firptr;
_complex* Inptr1;			//decimation FIR #1 variables
_complex* Qptr1;
_complex* Inptr2;			//decimation FIR #2 variables
_complex* Qptr2;
	m_DispTrig = FALSE;
	j=0;
	if( Freq != m_LastFreq )	//caller is changing center frequency
	{							//so recalc freq error value
		freqerror = freqerror + m_phzinc;	//current center freq inc
		m_phzinc = (PI2/m_Fs)*(double)Freq;	//new center freq inc
		if( UseAFC )
			freqerror = m_phzinc - freqerror;
		else
			freqerror = 0.0;
		m_LastFreq = Freq;
	}
	Inptr1 = m_pInPtr1;		//use local copies of member variables
	Qptr1 = m_pQue1;		// for better speed.
	Inptr2 = m_pInPtr2;
	Qptr2 = m_pQue2;
	for( i = 0; i<m_BlockSize; i++ )	// put new samples into Queue
	{
		if( --Inptr1 < Qptr1 )		//deal with FIR pointer wrap around
			Inptr1 = Qptr1+DEC3_LPFIR_LENGTH-1;
//Generate complex sample by mixing input sample with NCO's sin/cos
		Inptr1->x = pIn[i] * cos( vcophz );	//place in circular Queue
		Inptr1->y = pIn[i] * sin( vcophz );
		vcophz = vcophz + (m_phzinc + freqerror);
		if(vcophz > PI2)		//handle 2 Pi wrap around
			vcophz -= PI2;
//decimate by 3 filter
//  10uSec( 133MHz Pentium )
		if( ( (++m_SampCnt)%3 ) == 0 )	//calc first decimation filter every 3 samples
		{
			acc.x = 0.0;
			acc.y = 0.0;
			Firptr = Inptr1;
			Kptr = Dec3LPCoef;
			while( Kptr < (Dec3LPCoef + DEC3_LPFIR_LENGTH) )	//do the MAC's
			{
				acc.x += ( (Firptr->x)*(*Kptr) );
				acc.y += ( (Firptr->y)*(*Kptr++) );
				if( (++Firptr) >= Qptr1+DEC3_LPFIR_LENGTH ) //deal with wraparound
					Firptr = Qptr1;
			}
			if( --Inptr2 < Qptr2 )		//deal with FIR pointer wrap around
				Inptr2 = Qptr2+DEC3_LPFIR_LENGTH-1;
			Inptr2->x = acc.x;	//place in circular Queue
			Inptr2->y = acc.y;

//decimate by 3 filter
//  10uSec( 133MHz Pentium )
			if( (m_SampCnt%9) == 0 )	//calc second decimation filter every 9 samples
			{
				acc.x = 0.0;
				acc.y = 0.0;
				Firptr = Inptr2;
				Kptr = Dec3LPCoef;
				while( Kptr < (Dec3LPCoef + DEC3_LPFIR_LENGTH) )	//do the MAC's
				{
					acc.x += ( (Firptr->x)*(*Kptr) );
					acc.y += ( (Firptr->y)*(*(Kptr++)) );
					if( (++Firptr) >= Qptr2+DEC3_LPFIR_LENGTH ) //deal with wraparound
						Firptr = Qptr2;
				}
// here at Fs/9 = 612.5 Hz rate with latest sample in acc.
// Matched Filter the I and Q data and also a frequency error filter
//	filter puts filtered signals in m_FreqSignal and m_BitSignal.
				CalcBitFilter( acc );		// 25 uSec( 133MHz Pentium )
// Perform AGC operation
				m_SignalLevel = CalcAGC( m_FreqSignal );	// 9 uSec( 133MHz Pentium )
				if( m_DispTrig  )	//j ranges 0 to 2048/9 or 227
				{
					pDoc->m_pIQDispData[j].x = m_BitSignal.x;
					pDoc->m_pIQDispData[j].y = m_BitSignal.y;
					j++;
				}
// Calculate frequency error
				if( UseAFC )
				{
					freqerror = CalcFreqError(m_FreqSignal, pDoc);	// 10 uSec( 133MHz Pentium )
				}
				else
					freqerror = 0.0;
// Bit Timing synchronization
				if( SymbSync(m_BitSignal, pDoc) )	// 8 uSec( 133MHz Pentium )
				{
					DecodeSymb( m_BitSignal, pDoc);		//615uS(QPSK)
					m_DispTrig = TRUE;
				}
			}
		}
	}
	m_SampCnt = m_SampCnt%9;
	m_pInPtr1 = Inptr1;		// save position in circular delay line
	m_pInPtr2 = Inptr2;		// save position in circular delay line
	m_FreqError = freqerror;
	m_VcoPhz = vcophz;
	pDoc->m_ClkError = m_ClkError;
	return( (m_phzinc+freqerror)*m_Fs/PI2 );
}

//////////////////////////////////////////////////////////////////////
// Called at Fs/9 rate to calculate the symbol sync position
// Returns TRUE if at center of symbol.
// Sums up the energy at each sample time, averages it, and picks the
//   sample time with the highest energy content.
//////////////////////////////////////////////////////////////////////
BOOL CPskDet::SymbSync(_complex sample, CWinPSKDoc* pDoc)
{
BOOL Trigger;
double energy;
double max;
INT BitPos = m_BitPos;
	energy = (sample.x*sample.x) + (sample.y*sample.y);
	if( energy > 4.0)		//wait for AGC to settle down
		energy = 1.0;
	m_SyncAve[BitPos] = (1.0-1.0/100.0)*m_SyncAve[BitPos] + (1.0/100.0)*energy;
	if( BitPos == m_PkPos )	// see if at middle of symbol
	{
		Trigger = TRUE;
		pDoc->m_SyncHist[m_PkPos] = (INT)(10000.0*m_SyncAve[m_PkPos]);
	}
	else
	{
		Trigger = FALSE;
		pDoc->m_SyncHist[BitPos] = (INT)(7000.0*m_SyncAve[BitPos]);
	}
	if( BitPos == HALF_TBL[m_PkPos] )	//don't change pk pos until
		m_PkPos = m_NewPkPos;			// halfway into next bit.
	BitPos++;
	m_BitPhasePos += (m_BitPhaseInc);
	if( m_BitPhasePos >= Ts )
	{									// here every symbol time
		m_BitPhasePos -= Ts;			//keep phase bounded
		BitPos = 0;
		max = -1e10;
		for( INT i=0; i<19; i++)		//find maximum energy pk
		{
			energy = m_SyncAve[i];
			if( energy > max )
			{
				m_NewPkPos = i;
				max = energy;
			}
		}
		if( m_PkPos == m_LastPkPos+1 )	//calculate clock error
			m_ClkErrCounter++;
		else 
			if( m_PkPos == m_LastPkPos-1 )
				m_ClkErrCounter--;
		if( m_ClkErrTimer++ > 313 )	// every 10 seconds sample clk drift
		{
			m_ClkError = m_ClkErrCounter*180;
			m_ClkErrCounter = 0;
			m_ClkErrTimer = 0;
		}
		m_LastPkPos = m_PkPos;
	}
	m_BitPos = BitPos;
	return Trigger;
}


//////////////////////////////////////////////////////////////////////
//  Frequency error calculator
// calculates the derivative of the tan(I/Q).
// returns frequency error ~= .0034 per Hz error.
//////////////////////////////////////////////////////////////////////
double CPskDet::CalcFreqError( _complex IQ, CWinPSKDoc* pDoc )
{
#define FERRLPK (3000.0)	//freq error LP filter time constant
#define FERRLPG (4)			// freq filter Gain constant(1/4)
#define FCROSS (.048)		//freq where changes to using phase derived freq error
#define FEMIN (FCROSS/FERRLPG)	
double ferror;
	if(pDoc->m_UseNarrowAFC)	//just use phase error if after a TX session
	{							//  so rx freq doesn't wander too much
		if( m_QFreqError > .01 )//  clamp range to +/- 3 Hz
			m_QFreqError = .01;
		if( m_QFreqError < -.01 )
			m_QFreqError = -.01;
		m_FerAve = (1.0-1.0/FERRLPK)*m_FerAve + (1.0/(FERRLPG*FERRLPK))*m_QFreqError;
		return m_FerAve;
	}
	ferror = (IQ.x - m_z2.x) * m_z1.y - (IQ.y - m_z2.y) * m_z1.x;
	m_z2.x = m_z1.x;
	m_z2.y = m_z1.y;
	m_z1.x = IQ.x;
	m_z1.y = IQ.y;
	if( ferror > .15 )		//clamp range
		ferror = .15;
	if( ferror < -.15 )
		ferror = -.15;
// error at this point is abt .016 per Hz error
	if( (m_FerAve < -FEMIN) || (m_FerAve > FEMIN) )	//if error >FEMIN then use ferror
		m_FerAve = (1.0-1.0/FERRLPK)*m_FerAve + (1.0/(FERRLPG*FERRLPK))*ferror;
	else											//else use phase derived error
	{
		if( (m_QFreqError > -FCROSS) && (m_FerAve <= 0.0) )
			m_FerAve = (1.0-1.0/FERRLPK)*m_FerAve + (1.0/(FERRLPG*FERRLPK))*m_QFreqError;
		else
			m_FerAve = (1.0-1.0/FERRLPK)*m_FerAve + (1.0/(FERRLPG*FERRLPK))*ferror;
		if( (m_QFreqError < FCROSS) && (m_FerAve >= 0.0) )
			m_FerAve = (1.0-1.0/FERRLPK)*m_FerAve + (1.0/(FERRLPG*FERRLPK))*m_QFreqError;
		else
			m_FerAve = (1.0-1.0/FERRLPK)*m_FerAve + (1.0/(FERRLPG*FERRLPK))*ferror;
	}
	return m_FerAve;
}

//////////////////////////////////////////////////////////////////////
// Automatic gain control calculator
//////////////////////////////////////////////////////////////////////
double CPskDet::CalcAGC( _complex Samp)
{
double mag;
//	mag = fabs(Samp.x) + fabs(Samp.y);
	mag = sqrt(Samp.x*Samp.x + Samp.y*Samp.y);
	if( mag > m_AGCave )
		m_AGCave = (1.0-1.0/250.0)*m_AGCave + (1.0/250.0)*mag;
	else
		m_AGCave = (1.0-1.0/1000.0)*m_AGCave + (1.0/1000.0)*mag;
	if( m_AGCave >= 1.0 )	// divide signal by ave if not almost zero
	{
		m_BitSignal.x /= m_AGCave;
		m_BitSignal.y /= m_AGCave;
		m_FreqSignal.x /= m_AGCave;
		m_FreqSignal.y /= m_AGCave;
	}
	return(m_AGCave);
}

//////////////////////////////////////////////////////////////////////
//  Decode the new symbol sample
//////////////////////////////////////////////////////////////////////
void CPskDet::DecodeSymb(_complex newsamp, CWinPSKDoc* pDoc)
{
BOOL bit;
BYTE ch = 0;
BOOL GotChar = FALSE;
double angle;
	m_I3 = m_I2;		//form the multi delayed symbol samples
	m_Q3 = m_Q2;
	m_I2 = m_I1;
	m_Q2 = m_Q1;
	m_I1 = m_I0;
	m_Q1 = m_Q0;
	m_I0 = newsamp.x;
	m_Q0 = newsamp.y;
// calculates difference angle for QPSK, BPSK, and IQPSK decoding
//create vector whose angle is the difference angle by multiplying the
// current smaple by the complex conjugate of the previous sample.
//swap I and Q axis to keep away from  the +/-Pi discontinuity and
//  add Pi to make make range from 0 to 2Pi.
// 180 deg phase changes center at Pi/4
// 0 deg phase changes center at 3Pi/2
// +90 deg phase changes center at 2Pi or 0
// -90 deg phase changes center at Pi
//  if using lower sideband must flip sign of difference angle.
	if(pDoc->m_pSettings->m_ModType == CW_MODE )
		return;
	if((pDoc->m_pSettings->m_ModType == QPSK_MODE ) &&(pDoc->m_pSettings->m_UseLSB ))
		angle = (PI2/2) + atan2( (m_I1*m_I0 + m_Q1*m_Q0),
							(m_I0*m_Q1 - m_I1*m_Q0)); //IQPSK;
	else
		angle = (PI2/2) + atan2( (m_I1*m_I0 + m_Q1*m_Q0),
								(m_I1*m_Q0 - m_I0*m_Q1)); //QPSK or BPSK
	CalcQuality(pDoc, angle);
	if(pDoc->m_pSettings->m_ModType == BPSK_MODE )
		bit = GetBPSKSymb(pDoc->m_pSettings->m_UseMaxLikelyDet);
	else
	{
		if(pDoc->m_pSettings->m_UseMaxLikelyDet)
		{
			GetQPSKSymb();
			if(pDoc->m_pSettings->m_UseLSB )
			{	// Swap P90 and M90 symbols for lsb mode
				angle = m_QPSKprob[SYM_P90];
				m_QPSKprob[SYM_P90] = m_QPSKprob[SYM_M90];
				m_QPSKprob[SYM_M90] = angle;
			}
		}
		bit = ViterbiDecode( angle, pDoc->m_pSettings->m_UseMaxLikelyDet);
	}
	if( (bit==0) && m_LastBitZero )	//if character delimiter
	{
		if(m_BitAcc != 0 )
		{
			m_BitAcc >>= 2;				//get rid of last zero and one
			m_BitAcc &= 0x07FF;
			ch = m_VaricodeDecTbl[m_BitAcc];
			m_BitAcc = 0;
			GotChar = TRUE;
		}
	}
	else
	{
		m_BitAcc <<= 1;
		m_BitAcc |= bit;
		if(bit==0)
			m_LastBitZero = TRUE;
		else
			m_LastBitZero = FALSE;
	}
	if(GotChar && (ch!=0) && pDoc->m_SquelchOn)
	{
		::PostMessage(pDoc->m_hRcvEditCtrl,WM_USER_RCV_CHAR ,ch,RCV_FONT_ID);
	}
	GotChar = FALSE;
}

//////////////////////////////////////////////////////////////////////
// Calculate signal quality based on the statistics of the phase
//	difference angle.  The more dispersion of the "0" and "180" degree
//  phase shifts, the worse the signal quality.  This information is used
//  to activate the squelch control.  If 20 consecutive "180" degree shifts
//  occur, the squelch is forced on, and if 20 consecutive "0" degree
//  shifts occur, the squelch is forced off quickly.
//////////////////////////////////////////////////////////////////////
void CPskDet::CalcQuality( CWinPSKDoc* pDoc, double angle )
{
double temp;
	if( (angle >= PHZ_180_MIN) && (angle <= PHZ_180_MAX)  )
	{	//look +/-45 deg. around 180 deg.
		if((pDoc->m_pSettings->m_ModType == QPSK_MODE) && (pDoc->m_pSettings->m_UseLSB))
			temp = PI2/4.0 - angle;
		else
			temp = angle - PI2/4.0;
		m_QFreqError = (1.0/12.5)*temp;		//normalize same as freq error
		temp = fabs(temp);
		if( temp < m_DevAve)
			m_DevAve=  (1.0-1.0/50.0)*m_DevAve + (1.0/(50.0))*temp;
		else
			m_DevAve=  (1.0-1.0/100.0)*m_DevAve + (1.0/(100.0))*temp;
		m_OffCount = 0;
		if(m_OnCount > 20 )		// fast squelch counter
			m_DevAve = 0.1;
		else
			m_OnCount++;
	}
	else
		if( (angle >= PHZ_0_MIN) && (angle <= PHZ_0_MAX) )
		{		//look +/-45 deg around 0 deg.
			if((pDoc->m_pSettings->m_ModType == QPSK_MODE) && (pDoc->m_pSettings->m_UseLSB))
				temp = 3*PI2/4.0 - angle;
			else
				temp = angle - 3*PI2/4.0;
			m_QFreqError = (1.0/12.5)*temp;		//normalize same as freq error
			temp = fabs(temp);
			if( temp < m_DevAve)
				m_DevAve=  (1.0-1.0/50.0)*m_DevAve + (1.0/(50.0))*temp;
			else
				m_DevAve=  (1.0-1.0/100.0)*m_DevAve + (1.0/(100.0))*temp;
				m_OnCount = 0;
				if(m_OffCount > 20 )	// fast squelch counter
					m_DevAve = 0.4;
				else
					m_OffCount++;
		}
	pDoc->m_SquelchLevel = 100 - (INT)(250.0*m_DevAve);
	if(pDoc->m_SquelchLevel > pDoc->m_pSettings->m_SQThreshold )
		pDoc->m_SquelchOn = TRUE;
	else
		pDoc->m_SquelchOn = FALSE;
}

//////////////////////////////////////////////////////////////////////
// BIT FIR filters. A narrow matched(?) data filter for data
//                  and wider filter for the AFC/AGC functions
//////////////////////////////////////////////////////////////////////
void CPskDet::CalcBitFilter(_complex Samp)
{
_complex acc1;
_complex acc2;
const double* Kptr1;
const double* Kptr2;
_complex* Firptr;
_complex* Qptr;
_complex* Inptr;
	Inptr = m_pInPtr3;		//use local copies of member variables
	Qptr = m_pQue3;			// for better speed.
	if( --Inptr < Qptr )	//deal with LPFIR pointer wrap around
		Inptr = Qptr+BITFIR_LENGTH-1;
	Inptr->x = Samp.x;	//place in circular Queue
	Inptr->y = Samp.y;
	acc1.x = 0.0;
	acc1.y = 0.0;
	acc2.x = 0.0;
	acc2.y = 0.0;
	Firptr = Inptr;
	Kptr1 = FreqFirCoef;	//frequency error filter
	Kptr2 = BitFirCoef;	//bit data filter
	while( Kptr2 < (BitFirCoef + BITFIR_LENGTH) )	//do the MAC's
	{
		acc1.x += ( (Firptr->x)*(*Kptr1) );
		acc1.y += ( (Firptr->y)*(*Kptr1++) );
		acc2.x += ( (Firptr->x)*(*Kptr2) );
		acc2.y += ( (Firptr->y)*(*Kptr2++) );
		if( (++Firptr) >= (Qptr+BITFIR_LENGTH) ) //deal with wraparound
			Firptr = Qptr;
	}
	m_pInPtr3 = Inptr;		// save position in circular delay line
	m_FreqSignal.x = acc1.x;
	m_FreqSignal.y = acc1.y;
	m_BitSignal.x = acc2.x;
	m_BitSignal.y = acc2.y;
}

///////////////////////////////////////////////////////////////////////
// Experimental 4 chip BPSK decoder
///////////////////////////////////////////////////////////////////////
BOOL CPskDet::GetBPSKSymb(BOOL UseMaxLikely)
{
double v;
BOOL symb;
double max;
	if( UseMaxLikely )	//calc BPSK symbol over 4 chips
	{
		v = pow(m_I3 + m_I2 + m_I1 + m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_Q1 + m_Q0, 2 );   // v[1]   0 deg.
		max = v; symb = SYM_NOCHANGE;
		v = pow(m_I3 - m_I2 + m_I1 + m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_Q1 + m_Q0, 2 );   // v[2]   0 deg.
		if( v > max ) { max = v; symb = SYM_NOCHANGE;}
		v = pow(m_I3 + m_I2 + m_I1 + m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_Q1 + m_Q0, 2 );   // v[3]   0 deg.
		if( v > max ) { max = v; symb = SYM_NOCHANGE;}
		v = pow(m_I3 - m_I2 + m_I1 + m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_Q1 + m_Q0, 2 );   // v[4]   0 deg.
		if( v > max ) { max = v; symb = SYM_NOCHANGE;}
		v = pow(m_I3 + m_I2 - m_I1 - m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_Q1 - m_Q0, 2 );   // v[5]   0 deg.
		if( v > max ) { max = v; symb = SYM_NOCHANGE;}
		v = pow(m_I3 - m_I2 - m_I1 - m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_Q1 - m_Q0, 2 );   // v[6]   0 deg.
		if( v > max ) { max = v; symb = SYM_NOCHANGE;}
		v = pow(m_I3 + m_I2 - m_I1 - m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_Q1 - m_Q0, 2 );   // v[7]   0 deg.
		if( v > max ) { max = v; symb = SYM_NOCHANGE;}
		v = pow(m_I3 - m_I2 - m_I1 - m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_Q1 - m_Q0, 2 );   // v[8]   0 deg.
		if( v > max ) { max = v; symb = SYM_NOCHANGE;}
	//
		v = pow(m_I3 + m_I2 + m_I1 - m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_Q1 - m_Q0, 2 );   // v[9]   180 deg.
		if( v > max ) { max = v; symb = SYM_P180;}
		v = pow(m_I3 - m_I2 + m_I1 - m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_Q1 - m_Q0, 2 );   // v[10]   180 deg.
		if( v > max ) { max = v; symb = SYM_P180;}
		v = pow(m_I3 + m_I2 + m_I1 - m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_Q1 - m_Q0, 2 );   // v[11]   180 deg.
		if( v > max ) { max = v; symb = SYM_P180;}
		v = pow(m_I3 - m_I2 + m_I1 - m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_Q1 - m_Q0, 2 );   // v[12]   180 deg.
		if( v > max ) { max = v; symb = SYM_P180;}
		v = pow(m_I3 + m_I2 - m_I1 + m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_Q1 + m_Q0, 2 );   // v[13]   180 deg.
		if( v > max ) { max = v; symb = SYM_P180;}
		v = pow(m_I3 - m_I2 - m_I1 + m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_Q1 + m_Q0, 2 );   // v[14]   180 deg.
		if( v > max ) { max = v; symb = SYM_P180;}
		v = pow(m_I3 + m_I2 - m_I1 + m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_Q1 + m_Q0, 2 );   // v[15]   180 deg.
		if( v > max ) { max = v; symb = SYM_P180;}
		v = pow(m_I3 - m_I2 - m_I1 + m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_Q1 + m_Q0, 2 );   // v[16]   180 deg.
		if( v > max ) { max = v; symb = SYM_P180;}
		return !symb;		//bit is inverted symbol
	}
	else	//calc BPSK symbol over 2 chips
	{
		v = m_I1 * m_I0 +  m_Q1 * m_Q0;
		return (v>0.0);
	}
}

///////////////////////////////////////////////////////////////////////
// Soft-decision Viterbi decoder function.
///////////////////////////////////////////////////////////////////////
BOOL CPskDet::ViterbiDecode( double newangle,BOOL UseMaxLikely)
{
double pathdist[32];
double min;
INT bitestimates[32];
INT ones;
INT i;
const double* pAngleTbl;
	min = 1.0e100;		// make sure can find a minimum value
	if(UseMaxLikely)
	{
		for(i = 0; i < 32; i++)		// calculate all possible distances
		{							//lsb of 'i' is newest bit estimate
			pathdist[i] = m_SurvivorStates[i / 2].Pathdistance + 
							m_QPSKprob[ ConvolutionCodeTable[i] ];
			if(pathdist[i] < min)	// keep track of minimum distance
				min = pathdist[i];
			// shift in newest bit estimates
			bitestimates[i] = ((m_SurvivorStates[i / 2].BitEstimates) << 1) + (i & 1);
		}
	}
	else		//alternative soft decision method using angle differences
	{
		if( newangle >= PI2/2 )			//deal with ambiguity at +/- 2PI
			pAngleTbl = AngleTbl2;		// by using two different tables
		else
			pAngleTbl = AngleTbl1;
		for(i = 0; i < 32; i++)		// calculate all possible distances
		{							//lsb of 'i' is newest bit estimate
			pathdist[i] = m_SurvivorStates[i / 2].Pathdistance + 
					fabs(newangle - pAngleTbl[ ConvolutionCodeTable[i] ]);
			if(pathdist[i] < min)	// keep track of minimum distance
				min = pathdist[i];
			// shift in newest bit estimates
			bitestimates[i] = ((m_SurvivorStates[i / 2].BitEstimates) << 1) + (i & 1);
		}
	}
	for(i = 0; i < 16; i++)	//compare path lengths with the same end state
							// and keep only the smallest path in m_SurvivorStates[].
	{
		if(pathdist[i] < pathdist[16 + i])
		{
			m_SurvivorStates[i].Pathdistance = pathdist[i] - min;
			m_SurvivorStates[i].BitEstimates = bitestimates[i];
		}
		else
		{
			m_SurvivorStates[i].Pathdistance = pathdist[16 + i] - min;
			m_SurvivorStates[i].BitEstimates = bitestimates[16 + i];
		}
	}
	ones = 0;
	for(i = 0; i < 16; i++)		// find if more ones than zeros at bit 20 position
		ones += (m_SurvivorStates[i].BitEstimates&(1L << 20));
	if( ones == (8L << 20 ) )
		return ( rand() & 0x1000 );	//if a tie then guess
	else
		return(ones > (8L << 20) );	//else return most likely bit value
}

///////////////////////////////////////////////////////////////////////
// Experimental 4 chip QPSK decoder
///////////////////////////////////////////////////////////////////////
void CPskDet::GetQPSKSymb()
{
double v;
double max;
INT i;
//////////////   Calculate all 0 degree symbol possibilities
	max = pow(m_I3 - m_Q2 - m_Q1 - m_Q0, 2 ) + 
			pow(m_Q3 + m_I2 + m_I1 + m_I0, 2 );   // 0 deg.
	v = pow(m_I3 + m_Q2 - m_Q1 - m_Q0, 2 ) + 
			pow(m_Q3 - m_I2 + m_I1 + m_I0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 - m_Q1 - m_Q0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_I1 + m_I0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 - m_Q1 - m_Q0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_I1 + m_I0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 + m_Q1 + m_Q0, 2 ) + 
			pow(m_Q3 + m_I2 - m_I1 - m_I0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 + m_Q1 + m_Q0, 2 ) + 
			pow(m_Q3 - m_I2 - m_I1 - m_I0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 + m_Q1 + m_Q0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_I1 - m_I0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 + m_Q1 + m_Q0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_I1 - m_I0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 - m_I1 - m_I0, 2 ) + 
			pow(m_Q3 + m_I2 - m_Q1 - m_Q0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 - m_I1 - m_I0, 2 ) + 
			pow(m_Q3 - m_I2 - m_Q1 - m_Q0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 - m_I1 - m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_Q1 - m_Q0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 - m_I1 - m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_Q1 - m_Q0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 + m_I1 + m_I0, 2 ) + 
			pow(m_Q3 + m_I2 + m_Q1 + m_Q0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 + m_I1 + m_I0, 2 ) + 
			pow(m_Q3 - m_I2 + m_Q1 + m_Q0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 + m_I1 + m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_Q1 + m_Q0, 2 );   // 0 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 + m_I1 + m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_Q1 + m_Q0, 2 );   // 0 deg.
	if( v > max ) max = v;
	m_QPSKprob[SYM_NOCHANGE] = max;
//////////////   Calculate all 180 degree symbol possibilities
	max = pow(m_I3 - m_Q2 + m_Q1 - m_Q0, 2 ) + 
			pow(m_Q3 + m_I2 - m_I1 + m_I0, 2 );   // 180 deg.
	v = pow(m_I3 + m_Q2 + m_Q1 - m_Q0, 2 ) + 
			pow(m_Q3 - m_I2 - m_I1 + m_I0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 + m_Q1 - m_Q0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_I1 + m_I0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 + m_Q1 - m_Q0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_I1 + m_I0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 - m_Q1 + m_Q0, 2 ) + 
			pow(m_Q3 + m_I2 + m_I1 - m_I0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 - m_Q1 + m_Q0, 2 ) + 
			pow(m_Q3 - m_I2 + m_I1 - m_I0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 - m_Q1 + m_Q0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_I1 - m_I0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 - m_Q1 + m_Q0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_I1 - m_I0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 + m_I1 - m_I0, 2 ) + 
			pow(m_Q3 + m_I2 + m_Q1 - m_Q0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 + m_I1 - m_I0, 2 ) + 
			pow(m_Q3 - m_I2 + m_Q1 - m_Q0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 + m_I1 - m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_Q1 - m_Q0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 + m_I1 - m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_Q1 - m_Q0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 - m_I1 + m_I0, 2 ) + 
			pow(m_Q3 + m_I2 - m_Q1 + m_Q0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 - m_I1 + m_I0, 2 ) + 
			pow(m_Q3 - m_I2 - m_Q1 + m_Q0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 - m_I1 + m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_Q1 + m_Q0, 2 );   // 180 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 - m_I1 + m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_Q1 + m_Q0, 2 );   // 180 deg.
	if( v > max ) max = v;
	m_QPSKprob[SYM_P180] = max;
//////////////   Calculate all -90 degree symbol possibilities
	max = pow(m_I3 - m_Q2 - m_I1 - m_Q0, 2 ) + 
			pow(m_Q3 + m_I2 - m_Q1 + m_I0, 2 );   // -90 deg.
	v = pow(m_I3 + m_Q2 - m_I1 - m_Q0, 2 ) + 
			pow(m_Q3 - m_I2 - m_Q1 + m_I0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 - m_I1 - m_Q0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_Q1 + m_I0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 - m_I1 - m_Q0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_Q1 + m_I0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 + m_I1 + m_Q0, 2 ) + 
			pow(m_Q3 + m_I2 + m_Q1 - m_I0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 + m_I1 + m_Q0, 2 ) + 
			pow(m_Q3 - m_I2 + m_Q1 - m_I0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 + m_I1 + m_Q0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_Q1 - m_I0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 + m_I1 + m_Q0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_Q1 - m_I0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 + m_Q1 - m_I0, 2 ) + 
			pow(m_Q3 + m_I2 - m_I1 - m_Q0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 + m_Q1 - m_I0, 2 ) + 
			pow(m_Q3 - m_I2 - m_I1 - m_Q0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 + m_Q1 - m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_I1 - m_Q0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 + m_Q1 - m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_I1 - m_Q0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 - m_Q1 + m_I0, 2 ) + 
			pow(m_Q3 + m_I2 + m_I1 + m_Q0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 - m_Q1 + m_I0, 2 ) + 
			pow(m_Q3 - m_I2 + m_I1 + m_Q0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 - m_Q1 + m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_I1 + m_Q0, 2 );   // -90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 - m_Q1 + m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_I1 + m_Q0, 2 );   // -90 deg.
	if( v > max ) max = v;
	m_QPSKprob[SYM_M90] = max;

//////////////   Calculate all +90 degree symbol possibilities
	max = pow(m_I3 - m_Q2 + m_I1 - m_Q0, 2 ) + 
			pow(m_Q3 + m_I2 + m_Q1 + m_I0, 2 );   // 90 deg.
	v = pow(m_I3 + m_Q2 + m_I1 - m_Q0, 2 ) + 
			pow(m_Q3 - m_I2 + m_Q1 + m_I0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 + m_I1 - m_Q0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_Q1 + m_I0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 + m_I1 - m_Q0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_Q1 + m_I0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 - m_I1 + m_Q0, 2 ) + 
			pow(m_Q3 + m_I2 - m_Q1 - m_I0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 - m_I1 + m_Q0, 2 ) + 
			pow(m_Q3 - m_I2 - m_Q1 - m_I0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 - m_I1 + m_Q0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_Q1 - m_I0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 - m_I1 + m_Q0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_Q1 - m_I0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 - m_Q1 - m_I0, 2 ) + 
			pow(m_Q3 + m_I2 + m_I1 - m_Q0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 - m_Q1 - m_I0, 2 ) + 
			pow(m_Q3 - m_I2 + m_I1 - m_Q0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 - m_Q1 - m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 + m_I1 - m_Q0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 - m_Q1 - m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 + m_I1 - m_Q0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_Q2 + m_Q1 + m_I0, 2 ) + 
			pow(m_Q3 + m_I2 - m_I1 + m_Q0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_Q2 + m_Q1 + m_I0, 2 ) + 
			pow(m_Q3 - m_I2 - m_I1 + m_Q0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 - m_I2 + m_Q1 + m_I0, 2 ) + 
			pow(m_Q3 - m_Q2 - m_I1 + m_Q0, 2 );   // 90 deg.
	if( v > max ) max = v;
	v = pow(m_I3 + m_I2 + m_Q1 + m_I0, 2 ) + 
			pow(m_Q3 + m_Q2 - m_I1 + m_Q0, 2 );   // 90 deg.
	if( v > max ) max = v;
	m_QPSKprob[SYM_P90] = max;
//  normalize max to 1.0;
	max = -1e10;
	for(i=0; i<4;i++)
		if(m_QPSKprob[i] > max)
			max = m_QPSKprob[i];
	for(i=0; i<4;i++)
		m_QPSKprob[i] = -log10(m_QPSKprob[i]/max + 1E-100);
}
